/*									*/
/************************************************************************/
/* linux device driver for BNS 512^2 slm / pci system			*
 * (c) Matt Clark / GPL'd license					*/
/* assuming "no prefetchable memory" for both io and mem regions	*/

/* standard header files						*/
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>  
#include <asm/mtrr.h>

/* this driver will have NO text interface but will support /procfs	*/
#define VERSION         "v1.10 revision 21 banding problem fixed in bns_init_board() 5/03/04" 
/* ALLOW_HARD_RESET disables uniprocess locking- useful for "hardreset"	*/
//#define ALLOW_HARD_RESET

#include "slm.h"
#define DEBUG_PAUSE 250

/* Global vars	*/
static int slm_major=0;
static int slm_enable_mtrr=-1;

/* allow certain variables to be overwritten by the load process	*/
MODULE_PARM(slm_major,"i");
MODULE_PARM(slm_enable_mtrr,"i");

static struct proc_dir_entry *proc_file; 

/*	driver function prototypes					*/

static ssize_t slm_write(struct file *, const char *, size_t, loff_t *); 
static int slm_ioctl(struct inode *, struct file *, unsigned int , unsigned long); 
static int slm_open(struct inode *, struct file *);
static int slm_release(struct inode *, struct file *);
static void slm_int_handler(int, void *,struct pt_regs *);

/* utilities								*/
static inline void bns_select_image(u32, u32, u32, u32);
static void bns_ms_sleep(int);
static inline int bns_chk_sdram(void);
static inline void set_cntrl(u32);
static inline void up_cntrl(u32);
static inline void down_cntrl(u32);
static inline void set_image(int, int);
static inline void bns_enable_write(void);
static inline void bns_disable_write(void);
static inline void bns_disable_DPEN(void);
static inline void bns_enable_DPEN(void);
static inline void bns_enable_ints(void);
static inline void bns_disable_ints(void);
static int bns_init_board(void);
static int bns_power_on(void);
static int bns_power_off(void);
static int bns_proc_read(char *, char **, off_t, int, int *, void *);
static int bns_init_sdram(void);
static inline int bns_check_RRDY(void);
static int bns_probe_epi(void);
static int bns_probe_memory_board(void);
static void bns_probe_plx_int(void);

/* fops defines the standard interface for char devices			*
 * this driver will use write to access memory in the first instance	*
 * ioctl to perform most "user" functions				*
 * and later on I will implement mmap for raw IO			*
 * I will support 2 write modes- direct write and "scrambling" write	*
 * which will change the byte order appropriately 			*
 * I will restrict writing to row-at-a-time and include a "set row"	*
 * ioctl (which I can't see being used					*/
 
struct file_operations slm_fops = {
//      read:		slm_read,
	write:		slm_write,
	ioctl:		slm_ioctl,
	open:		slm_open,
	release:	slm_release,
//      mmap:		slm_mmap,
	};

// local version of control register rather than reading off of the board
static u32 control;




/************************************************************************/
/* INIT									*/
/* note kernel 2.2 style init definition---				*/

int init_module(void){
struct pci_dev *dev=NULL;
int	err;
char 	int_n;

        printk("\nBoulder Non Linear 512^2 slm system driver %s (c)mc\n\n",VERSION); 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
/* they shouldn't compile in any event- matt oct 2001			*/
	printk("slm: init: kernels earlier than 2.4.0 are not supported\n");
	return -ENODEV;
#endif

	if(!(dev=pci_find_device(BNS_VENDOR_ID,BNS_DEVICE_ID,dev))){
		printk("slm: init: no BNS slm card found in this machine\n");
		return -ENODEV;
		}
	/* this should never happen- trapping it for coding errors only */
	if(dev==NULL){
		printk("slm: init: card found but dev==NULL\n");
		return -ENODEV;
		}
/* At this point we <should> have a unique id for the card, however BNS	*
 * don't have a unique vendor/device number pair at this time (oct 2001	*
 * so we may have to install additional card probing here if required	*/
	
/* get the base ids for slm_io and slm_mem				*/
/* please note that at present I don't know which region is which	*/
/* trap errors because of incomplete information			*/

/* wake the card up							*/
	if((err=pci_enable_device(dev))!=0){
		/* we are in deep trouble here if this fails- matt	*/
		printk("slm: init: pci_enable failed with %d exiting\n",err);
		return -ENODEV;
		}

/* get the interrupt line #						*/
	if((err=pci_read_config_byte(dev,PCI_INTERRUPT_LINE,&int_n))!=0){
		/* There is no interrupt for this card!			*/
		printk("slm: init: no interrupt line found for card- exiting\n");
		return -ENODEV;
		}
	slm.interrupt=int_n;
/* map the resources to virual memory					*/
	slm.plx_chip=ioremap(pci_resource_start(dev,BNS_SLM_PLX)&PCI_BASE_ADDRESS_MEM_MASK,BNS_SLM_PLX_SIZE);
	if(slm.plx_chip==NULL){
		printk("slm: init: Unable to map plx chip space  (region %d, %x bytes) for device\n",BNS_SLM_IO,BNS_SLM_IO_SIZE);
		err= -ENODEV;
		goto init_error_1;
		}
	
	slm.io=ioremap(pci_resource_start(dev,BNS_SLM_IO)&PCI_BASE_ADDRESS_MEM_MASK,BNS_SLM_IO_SIZE);
	if(slm.io==NULL){
		printk("slm: init: Unable to map iospace  (region %d, %x bytes) for device\n",BNS_SLM_IO,BNS_SLM_IO_SIZE);
		err= -ENODEV;
		goto init_error_2;
		}
/* get the physical address and sort it for MTRR			*/
	slm.phys_mem=pci_resource_start(dev,BNS_SLM_MEM);
	slm.mem=ioremap(slm.phys_mem&PCI_BASE_ADDRESS_MEM_MASK,BNS_SLM_MEM_SIZE);
	if(slm.mem==NULL){
		printk("slm: init: Unable to map memory  (region %d, %x bytes) for device\n",BNS_SLM_MEM,BNS_SLM_MEM_SIZE);
		err= -ENODEV;
		goto init_error_3;
		}
	printk("slm: init: IO regions\n");
	printk("slm: init: plx chip =%lx\n",(unsigned long)slm.plx_chip);
	printk("slm: init: io       =%lx\n",(unsigned long)slm.io);
	printk("slm: init: memory   =%lx\n",(unsigned long)slm.mem);
/* mtrr region - */
	if(slm_enable_mtrr>=0){
		printk("slm: init: mtrr option enabled- trying region %lx\n",slm.phys_mem);
		slm_enable_mtrr=mtrr_add(slm.phys_mem,BNS_SLM_MEM_SIZE,MTRR_TYPE_WRCOMB,1);
		/* check and report */
		if(slm_enable_mtrr<0){
			printk("slm: init: mtrr: unable to set write combining for slm memory :(\n");
			}
		else{
			printk("slm: init: mtrr Write combining enabled (mtrr handle=%d)\n",slm_enable_mtrr);
			}
		}
	else{
		printk("slm: init: mtrr option not enabled\n");
		}

/* announce the slm 							*/
	printk("slm: init: BNS 512^ device  card found- probing for additional info\n");
/* perform additional probing here - first enable the borad		*/

/* this is the first port of call, although not documented I will start with 	*
 * setting cntrl to 0- I can't see the need for any 1's although the rrdy bit	*
 * will have to be set sometime- matt						*/
	set_cntrl(0);
	if(readl(slm.io+BNS_CONTROL) != 0){
		printk("slm: init: control word inconsistent\n");
		}
	bns_init_board();
	/* check the ID of the 2port					*/
	if((err=bns_probe_epi())!=0){
			printk("slm: init: bns_probe_epi failed\n");
			goto init_error_4;
			}

	/* find out how much memory is on board				*/
	if((err=bns_probe_memory_board())!=0){
			printk("slm: init: bns_probe_memory_board failed\n");
			goto init_error_4;
			}  

/* get and check the sdram IDs						*/
	if((err=bns_init_sdram())!=0){
#ifndef BNS_IGNORE_MEM_ERROR		
                        printk("slm: init: SDRAM fatal error: cleaning and exiting\n");
			goto init_error_4;
#endif
			}
/* we claim the interrupt line during the entire module install because	*
 * device can operate async even when not open <AND> because it must 	*
 * never fail								*/
	printk("slm: init: about to probe plx interrupt space\n");
	bns_probe_plx_int();
	printk("slm: init: installing interrupt handler on line %d\n",slm.interrupt);
	err=request_irq(slm.interrupt,slm_int_handler,SA_SHIRQ,IRQ_ID,(void *)&slm);
	if(err!=0){
                        printk("slm: init: Could not register interrupt %d, cleaning and exiting\n",slm.interrupt);
			goto init_error_4;
			}
	/* All h/w probing to be over by here 				*/

	/* init global structures 1st 					*/
	slm.image_list=(BNS_irl_object *)kmalloc(sizeof(BNS_irl_object)*slm.max_images,GFP_KERNEL);
	if(slm.image_list==NULL){
			printk("slm: init: unable to kmalloc %d bytes, cleaning up and exiting\n",sizeof(BNS_irl_object)*slm.max_images);
			err=-ENOMEM;
			goto init_error_5;
			}	
/* now register the char device 					*/
	err=register_chrdev(slm_major,SLM_MOD_NAME,&slm_fops);
	if (err<0){
		printk("slm: init: unable to register char dev using major=%d\n",slm_major);
		goto init_error_5;
		}
	slm_major=err; /* not an error if +ve*/
	
	printk("slm: init: device major=%d\n",slm_major); 
	printk("slm: init: installing /proc \n"); 
	proc_file=create_proc_read_entry(SLM_MOD_NAME,0,NULL,bns_proc_read,NULL);
	if(!proc_file){
		printk("slm: init: error installing /proc file- sorry -non fatal\n");
		}
	printk("slm: init: installation complete\n");
	bns_ms_sleep(DEBUG_PAUSE);
	

	return 0;
/* GOTOs for errors "the only place for gotos" -Linus */
	init_error_5:
		free_irq(slm.interrupt,(void *)&slm);
		printk("slm: init: interrupt released\n");
	init_error_4:
		if(slm_enable_mtrr>=0){
			if(slm_enable_mtrr==mtrr_del(slm_enable_mtrr,(unsigned long)slm.phys_mem,BNS_SLM_MEM_SIZE))
				printk("slm: mtrr: released ok\n");
			else printk("slm: mtrr: there was a problem\n");
			}
		if(slm.mem!=NULL)iounmap(slm.mem);
		printk("slm: init: memory region released\n");
	init_error_3:
		if(slm.io!=NULL)iounmap(slm.io);
		printk("slm: init: io region released\n");
	init_error_2:
		if(slm.plx_chip!=NULL)iounmap(slm.plx_chip);
		printk("slm: init: PLX chipset region released\n");
	init_error_1:
		printk("slm: init: Failed to init- no module loaded\n");	
	return err;
	}


/************************************************************************/
/* CLEAN								*/
void cleanup_module(void){
int res;
	/* free alloc'd resources, set to NULL in case of reentrance	*/
	printk("slm: cleanup_module: Matt Clark oct 2001- device %d\n",slm_major);

	/* ALWAYS do io first - we need to clear up the power etc here	*/
	if(slm.io!=NULL){
		printk("slm: cleanup: enable writes\n");
		bns_ms_sleep(DEBUG_PAUSE);
		bns_enable_write();
		printk("slm: cleanup: free_irq\n");
		bns_ms_sleep(DEBUG_PAUSE);
		free_irq(slm.interrupt,(void *)&slm);
		printk("slm: cleanup: bns_power_off\n");
		bns_ms_sleep(DEBUG_PAUSE);
		bns_power_off();
		printk("slm: cleanup: unmap io\n");
		bns_ms_sleep(DEBUG_PAUSE);
		iounmap(slm.io);
		slm.io=NULL;
		}
	if(slm.plx_chip!=NULL){
		printk("slm: cleanup: unmap plx chip\n");
		bns_ms_sleep(DEBUG_PAUSE);
		iounmap(slm.plx_chip);
		slm.plx_chip=NULL;
		}
	if(slm.mem!=NULL){
		// release the write combining region first
		if(slm_enable_mtrr>=0){
			printk("slm: mtrr: releasing write-combining region\n");
			if(slm_enable_mtrr==mtrr_del(slm_enable_mtrr,(unsigned long)slm.phys_mem,BNS_SLM_MEM_SIZE)){
				printk("slm: mtrr: released ok\n");
				}
			else{
				printk("slm: mtrr: there was a problem\n");
				}
			}
		printk("slm: cleanup: unmap mem\n");
		bns_ms_sleep(DEBUG_PAUSE);
		iounmap(slm.mem);
		slm.mem=NULL;
		}
	if(slm.image_list!=NULL){
		printk("slm: cleanup: free image_list\n");
		bns_ms_sleep(DEBUG_PAUSE);
		kfree(slm.image_list);
		slm.image_list=NULL;
		}
	if(proc_file){
		printk("slm: cleanup: remove proc\n");
		bns_ms_sleep(DEBUG_PAUSE);
		remove_proc_entry(SLM_MOD_NAME,NULL);
		}
	res=unregister_chrdev(slm_major,SLM_MOD_NAME);
	}


/************************************************************************/
/* IOCTL								*/
static int slm_ioctl(struct inode * inode, struct file *file, unsigned int cmd, unsigned long arg){
int temp,i;
	switch(cmd){

#ifdef ALLOW_HARD_RESET
		case BNS_HARDRESET:
			/* won't work unless ALLOW-HARD-RESET is defined		*/
			/* resets the module count to 1 so that the 			*
			 * module can be removed if crashed				*/
			printk("slm: ioctl: HARDRESET: attempting mod_use=1\n");
						/* pull it to 0 			*/
			while(MOD_IN_USE)MOD_DEC_USE_COUNT;
			/* inc once for this open 					*/
			MOD_INC_USE_COUNT;
			/* set locked to FALSE 						*/
			slm.locked=FALSE;
			printk("slm: Done\n");
			break;
#endif

		case BNS_DISPLAY_IMAGE:
			/* takes a image AND displays it, and disables ints		*/
			if(arg>slm.max_images)return -EINVAL;
			slm.image_mode=BNS_SLM_IMAGE;
			bns_select_image((int)arg,
				 slm.periods,
				 slm.lasera,
				 slm.laserb);
			if(bns_check_RRDY()){
				slm.int_errs++;
				/* reset the RRDY error done in check_RRDY*/
				}
			/* enable DPEN */
			PRINT_DEBUG("slm: ioctl: BNS_DISPLAY_IMAGE slm.cur_image=%d, mode=slm_image\n",slm.cur_image);
			break;
		case BNS_UPLOAD_IMAGE:
			/* sets the upload image and sets the memory pointers to 0 	*/
			if(arg>slm.max_images)return -EINVAL;
			slm.row=0;
			slm.mem_image=(int)arg;
			set_image(slm.mem_image,slm.cur_image);
			PRINT_DEBUG("slm: ioctl: BNS_UPLOAD_IMAGE: upload to image %d, row=0\n",slm.mem_image);
			break;
		case BNS_SET_ROW:
			/* sets the current row # */
			if(arg>BNS_MAX_ROWS)return -EINVAL;
			slm.row=(int)arg;	
			PRINT_DEBUG("slm: ioctl: BNS_SET_ROW: row=%d\n",slm.row);
		case BNS_POWER_ON:
			/* go through power up sequence and disable ints		*/
			bns_power_on();
			PRINT_DEBUG("slm: ioctl: BNS_POWER_ON\n");
			break;
		case BNS_POWER_OFF:
			/* go through power down sequence and disable ints		*/
			bns_power_off();
			PRINT_DEBUG("slm: ioctl: PNS_POWER_OFF\n");
			break;
		case BNS_SET_LIST_LENGTH:
// XXX MATT matt race condition here between using slm.n_list for
// uploads and going through the list
// need to keep an internal copy for inturrpts and
// update at BNS_START_LIST - matt march 3
			if(arg>slm.max_images)return -EINVAL;
			slm.n_list=(int)arg;
			PRINT_DEBUG("slm: ioctl: BNS_SET_LIST_LENGTH: row=%d\n",slm.n_list);
			break;
		case BNS_SET_LIST:
			/* uploads slm.n_list of image data and checks it		*/	
			copy_from_user((void *)slm.image_list,(void *)arg,slm.n_list*sizeof(BNS_irl_object));	
			/* check the values	*/
			for(i=0;i<slm.n_list;i++){
				if(slm.image_list[i].image<0)		slm.image_list[i].image=0;
				if(slm.image_list[i].image>1024)	slm.image_list[i].image=1024;
				if(slm.image_list[i].periods<6)		slm.image_list[i].periods=6;
				if(slm.image_list[i].periods>0xffff)	slm.image_list[i].periods=0xffff;
				}
				
			PRINT_DEBUG("slm: ioctl: BNS_SET_LIST:\n");
			break;
		case BNS_START_LIST:
			/* sets mode=SLM_LIST and enables ints*/
			/* takes the arg as the start image				*/
			if(arg>slm.n_list)return -EINVAL;

			slm.list_pos=(int)arg;
			slm.image_mode=BNS_SLM_LIST;	
			/* start the list going this way - just disable writes again	*
			 * this opens DPEN and starts interrupts			*/
			bns_disable_write();
			PRINT_DEBUG("slm: ioctl: BNS_START_LIST: at=%d\n",slm.list_pos);
			break;
		case BNS_STOP_LIST:
			/* sets mode=SLM_LIST and disable ints, image state unknown	*/
			/* use BNS_DISPLAY_IMAGE to park to a particular image		*/
			/* use BNS_SYSTEM_DISABLE to turn off the system		*/
			slm.image_mode=BNS_SLM_OFF;	
			/* stop the list going this way - just enable writes again	*
			 * this opens DPEN and starts interrupts			*/
			bns_enable_write();
			PRINT_DEBUG("slm: ioctl: BNS_STOP_LIST:\n");
			break;
		case BNS_GET_IMAGE:
			/* returns the current image number 				*/
			PRINT_DEBUG("slm: ioctl: BNS_GET_IMAGE:\n");
			return slm.cur_image;
			break;
		case BNS_SET_LASERA:
			temp=(int)arg;
//			if(temp<0 || temp>255){
//				return -EINVAL;
//				}
			slm.lasera=temp;
			//hacked by sds to make this take immediate effect
		        writel(slm.lasera,slm.io+BNS_LASERA);
			
			break;
		case BNS_SET_LASERB:
			temp=(int)arg;
//			if(temp<0 || temp>255){
//				return -EINVAL;
//				}
			slm.laserb=temp;			
			//hacked by sds to make this take immediate effect
		        writel(slm.laserb,slm.io+BNS_LASERB);
			break;
		case BNS_SET_PERIODS:
			slm.periods=(BNS_IMAGE_PERIODS & arg);			
			if(slm.periods<6){
				slm.periods=6;
				return -EINVAL;
				}
			writel(slm.periods,slm.io+BNS_PERIODS);
			break;

/* while enable_write and disable_DPEN do differ as far as the user is concerned they 	*/
/* are the same- in any case the differences are technical and not the users concern	*/

		case BNS_ENABLE_WRITE:
		/* case BNS_DISBALE_DPEN: */
			bns_enable_write();
			break;
		case BNS_DISABLE_WRITE:
		/* case BNS_ENABLE_DPEN: */
			bns_disable_write();
			break;

		case BNS_SET_DPEN_MODE:
			temp=(int)arg;
			if(temp<0 || temp>BNS_DPEN_MODE_MAX){
				printk("slm: illegal DPEN mode\n");
				return -EINVAL; 
				}
			slm.DPEN_mode=temp;
			break;
		case BNS_SET_DPEN_DELAY:
			temp=(int)arg;
			if(temp<BNS_DEFAULT_DPEN_DELAY){
				PRINT_DEBUG("slm: warning dangerous BNS_SET_DPEN_DELAY set (%d, min safe=%d)\n",temp,BNS_DEFAULT_DPEN_DELAY);
				bns_ms_sleep(DEBUG_PAUSE);
				}
			slm.DPEN_delay=temp;
			break;
		//default: aparently deprecated ?? gcc 3.2.2!
			/* during development report other events, after dev do nowt	*/	
		}
	return 0;
	}	


/****************************************************************************************/
/* WRITE										*/
/* A big function here- should implement two modes PERFORM_SCRAMBLE / PERFORM_RAW	*
 * The function only consumes mulitples of ROW_SIZE and throws an error otherwise	*
 * I could implement buffering for lower write sizes but I won't- this makes the	*
 * relationship between scrambled and unscrambled data and transfer sizes explicit	*/

/* consume data row at a time, writing it CHUNK at a time				*/
/* turns out if these are ~512 bytes then the overheads are minimal on substitute	*/
/* scrambling overheads on substitute are minimal also					*/

/* because we have written this over and over to try to get some decent speed out of it	*
 * it became difficult to read- hence this slightly dumb "easier to read" version	*
 * should have no speed issues- all the time appears to be spent in copy_from_usr and	*
 * memcpy_toio- the overheads for calling both these appears insignificant as does the	*
 * scrambling overhead	- matt nov 2001							*/
/* I suspect the sluggish nature of this to be IO bandwidth on substitute rather than	*
 * a coding issue 9/11/01								*/
/* certainly runs a lot faster on a decnt machine ("blanket"/athlon 2600 cf 		*
 * "substitute" p500) matt aug 2003							*/

static ssize_t slm_write(struct file * file,const  char * buf, size_t count, loff_t *ppos){

/* i	=index to the rows of data passed						*/
/* mpos	=index to the correct place in PCI memory					*/
/* bpos	=index to the correct place in buffer memory					*/
/* col 	=index to col within row for scrambling						*/
/* rowb	=row buffer,	scrb=scrambled row buffer					*/

/* profiling numbers 							*/
#ifdef BNS_PROFILE
unsigned long prof_1=0,prof_2=0,prof_3=0,prof_4=0; 
#endif


int rows,i,col,bpos,mpos; 
unsigned char rowb[BNS_ROW_SIZE],scrb[BNS_ROW_SIZE];

/* lock out writing during DPEN enable	*/


/* On older cards the machines crashed if written too during dual port transfers	*
 * if the card is one of these the this code prevents this event			*/
#ifdef BNS_2PORT_LOCK_OUT
	if(slm.DPEN_enable==TRUE){
		return -EBUSY;
		}
#endif

	/* first check the transfer size is a * of the ROW_SIZE				*/
	rows=count/BNS_ROW_SIZE;
	if((rows*BNS_ROW_SIZE) != count){
		/* not a mulitple of rows						*/
		PRINT_DEBUG("slm: write: transfer request not a multiple of BNS_ROW_SIZE\n");
		return -EINVAL;
		}

	/* check we didn't over run the frame						*/
	if(slm.row>=BNS_N_ROWS){
		/* we have filled the image already!					*/
		PRINT_DEBUG("slm: write: row>= BNS_N_ROWS\n");
		return -EINVAL;
		}

	/* now data IS a multiple of BNS_ROW_SIZE					*/
	/* grab and process the lump one by one						*/
	

/* if the sdram controllers are dodgy then check check away				*/
#ifdef BNS_DODGY_SDRAM
	if(bns_chk_sdram()!=BNS_SDRAM_OK){
		printk("SLM: sdram err in write\n");
		return -EIO;
		}
#endif

	/* main loop- attempting to make it easy to read- matt nov 2001			*/

	bpos=0;
	mpos=slm.row*BNS_ROW_SIZE;

	for(i=0;(i<rows) && (slm.row<BNS_N_ROWS); i++,slm.row++){
		TIC(prof_1); /* store clock ticks */
		/* consume a row from the buffer					*/
		copy_from_user(rowb,buf+bpos,BNS_ROW_SIZE);
		TIC(prof_2); /* store clock ticks */
		if(slm.transfer_mode==PERFORM_SCRAMBLE){
			/* scramble data- note only the col pos is significant		*/
			for(col=0;col<BNS_ROW_SIZE;col++){
				scrb[col]=rowb[BNS_ADD_TO_SLM(col)];	
				}
			TIC(prof_3); /* store clock ticks */
			memcpy_toio(slm.mem+mpos,scrb,BNS_ROW_SIZE);
			TIC(prof_4); /* store clock ticks */
			}
		else	memcpy_toio(slm.mem+mpos,rowb,BNS_ROW_SIZE); 
	
		mpos+=BNS_ROW_SIZE;
		bpos+=BNS_ROW_SIZE;
		}
	

	/* keep a count of the number of transfers for /proc/				*/
	if(slm.row==BNS_N_ROWS){
		slm.uploaded++;
		slm.row=0;
		}
#ifdef BNS_PROFILE
	/* profile that the critical regions */
	slm.sec_copy_from=prof_2-prof_1;
	slm.sec_scramble=prof_3-prof_1;
	slm.sec_pci_write=prof_4-prof_1;
#endif


        return count;
        }

/****************************************************************************************/
/* INTERRUPT 										*/
/* this increments slm.list_pos and pushes the next image onto the card			*/
/* it is a "slow" handler by linux defs, however it is a very fast operation and is	*
 * managed without bottom halves or taskets						*/
/* ignore the arguments- we don't need them						*/
/* DPEN should always be open- during host->card uploads DPEN and interrupts are 	*
 * disabled so DPEN is always enabled elsewhere						*/

static void slm_int_handler(int irq, void *dev_id,struct pt_regs *regs){
u32	temp;

/* probe to see if this is ours								*/
	temp=readl(slm.plx_chip+PLX_INTERRUPT);
	if(!(temp&PLX_INT_LINT1_STAT)){
		/* its not ours inc not ints						*/
		slm.not_ints++;
		return;
		}
	/* it is ours									*/	
	slm.ints++;

/* 1st acknowledge the interrupt							*/
/* exact order here unclear- but between twang and release do something to give the	*
 * board a chance - even if it doesn't need it- matt					*/
	/* (a) twang the board								*/
	writel(temp|PLX_INT_LET_CLR1,slm.plx_chip+PLX_INTERRUPT);
	/* (b) release BNS local bus							*/
	writel(0,slm.io+BNS_CLRINT);
	/* (c) release the board							*/
	writel(temp&~PLX_INT_LET_CLR1,slm.plx_chip+PLX_INTERRUPT);
	
/* now down to business- get the next image						*/
	if(slm.list_pos>=slm.n_list)slm.list_pos=0;
/* push the image onto the card								*/
	bns_select_image(slm.image_list[slm.list_pos].image,
			 slm.image_list[slm.list_pos].periods,
			 slm.lasera,
			 slm.laserb);
	if(bns_check_RRDY()){
		slm.int_errs++;
		/* reset the RRDY error done in check_RRDY*/
		}
	/* bug fix for updating the cur_image for get_image correctly matt July 2003 */
	slm.cur_image=slm.image_list[slm.list_pos].image;
/* done! */
	slm.list_pos++;
	slm.ints++;
	}


/****************************************************************************************/
/* /PROC filesystem									*/
/* implement a read only /proc filesystem that gives useful info about the card		*/
/* fill 1 page max- so ignore start, offset and we don't need data -no private data	*/

static int bns_proc_read(char * buf, char **start, off_t offset, int count, int *eof, void *data){
int w=0;
	/* force and eof for "normal library functions"	*/
	if(offset!=0){
		*eof=1;
		return 0;
		}
	if(count<70){
		/* not enough space for a warning*/
		return 0;
		}
	if((count>=70)&&(count<1024)){
		printk("Sorry: due to a lazy programmer the /proc buffer size must be > 1k\n");
		}
//	size=count<PAGE_SIZE?count:PAGE_SIZE;
/* now the idea for this /proc entry is to give basic slm card info for the user	*/
        w+=sprintf(buf+w,"Boulder Non Linear 512^2 slm system driver %s (c)mc\n\n",VERSION);  
	if(slm.io!=NULL) 	w+=sprintf(buf+w,"IO is mapped, ");
	else		 	w+=sprintf(buf+w,"IO is unmapped, ");
	if(slm.mem!=NULL) 	w+=sprintf(buf+w,"MEM is mapped\n");
	else		 	w+=sprintf(buf+w,"MEM is unmapped\n");
	w+=sprintf(buf+w,"IRQ line #            %d\n",slm.interrupt);
	w+=sprintf(buf+w,"mem board ID          %x\n",slm.mem_board_id);
	w+=sprintf(buf+w,"epi ID                %x",slm.epi_id);
	if(slm.epi_id==BNS_EPI_NOT_PRESENT)
		w+=sprintf(buf+w," (not present)\n");
	else
		w+=sprintf(buf+w," (present)\n");

	w+=sprintf(buf+w,"SDRAM controller ID   %x ",slm.sdram_id);
	if(slm.sdram_id==BNS_SDRAM_FAIL)
		w+=sprintf(buf+w,"(failed)\n");
	else
		w+=sprintf(buf+w,"(good)\n");
	
	w+=sprintf(buf+w,"card/driver supports  %d images\n\n",slm.max_images);

	if(slm.locked==TRUE)	w+=sprintf(buf+w,"the driver is locked (in use), ");
	else			w+=sprintf(buf+w,"the driver is unlocked, ");
	
	w+=sprintf(buf+w,"mode is ");
	switch(slm.image_mode){
		case BNS_SLM_OFF:
			w+=sprintf(buf+w,"SLM_OFF\n");
		break;		
		case BNS_SLM_IMAGE:
			w+=sprintf(buf+w,"SLM_IMAGE\n");
		break;		
		case BNS_SLM_LIST:
			w+=sprintf(buf+w,"SLM_LIST\n");
		break;		
		}
	if(slm.image_mode==BNS_SLM_IMAGE){
		w+=sprintf(buf+w,"image # on slm	 %d\n",slm.cur_image);
		}
	if(slm.image_mode==BNS_SLM_LIST){
		w+=sprintf(buf+w,"list # on slm		 %d/%d (image # %d)\n",slm.list_pos, slm.n_list, slm.image_list[slm.list_pos].image);
		}

	w+=sprintf(buf+w,"\nupload is to image %d (row %d), ",slm.mem_image,slm.row);
	if(slm.transfer_mode==PERFORM_SCRAMBLE)	
		w+=sprintf(buf+w,"scrambling image in driver\n");
	else	w+=sprintf(buf+w,"transfer images raw\n");
	
	w+=sprintf(buf+w,"\nDIAGNOSTICS:\n");
			
	w+=sprintf(buf+w,"power status : ");
	if(slm.power_status&BNS_POWER_DAC) 	w+=sprintf(buf+w,"DAC(on) : ");
	else					w+=sprintf(buf+w,"DAC(off) : ");
	if(slm.power_status&BNS_POWER_OPAMP) 	w+=sprintf(buf+w,"OPAMP(on) : ");
	else					w+=sprintf(buf+w,"OPAMP(off) : ");
	if(slm.power_status&BNS_POWER_SLM) 	w+=sprintf(buf+w,"SLM(on) : ");
	else					w+=sprintf(buf+w,"SLM(off) : ");
	if(slm.power_status&BNS_POWER_VOK) 	w+=sprintf(buf+w,"3.3v(on)\n");
	else					w+=sprintf(buf+w,"3.3v(off)\n");
	w+=sprintf(buf+w,"laser A              %d\n",slm.lasera);
	w+=sprintf(buf+w,"laser B              %d\n",slm.laserb);
	w+=sprintf(buf+w,"image uploads        %d\n",slm.uploaded);
	w+=sprintf(buf+w,"SDRAM control errs   %d\n",slm.mem_errs);
	w+=sprintf(buf+w,"SDRAM S2 0x89 errs   %d\n",slm.mem_err0x89);
	w+=sprintf(buf+w,"interrupts serviced  %d\n",slm.ints);
	w+=sprintf(buf+w,"interrupts not ours  %d\n",slm.not_ints);
	w+=sprintf(buf+w,"int errs (RRDY_ERR)  %d\n",slm.int_errs);
	w+=sprintf(buf+w,"images displayed     %d\n",slm.displayed);
#ifdef BNS_PROFILE
	w+=sprintf(buf+w,"\nProfiling info\n");
	w+=sprintf(buf+w,"section copy_from    %d\n",slm.sec_copy_from);
	w+=sprintf(buf+w,"section scramble     %d\n",slm.sec_scramble);
	w+=sprintf(buf+w,"section pci_write    %d\n",slm.sec_pci_write);
#else
	w+=sprintf(buf+w,"\nNo profile infor\n");
#endif
	
	
	/* w+=sprintf(buf+w,"count %d /proc/ fs size %d+20XXXX\n",count,w);*/
	*eof=1;
	return w;
	}	


/****************************************************************************************/
/* OPEN											*/
/* in this case- global device structure init'd at driver install this doen't do much	*/
/* simply enforce single open access */
/* global locking here prevents ioctl */
/* if multiple cards are needed- use minor to reference their structures		*/

static int slm_open(struct inode *inode, struct file *file){
	 //XXXX matt disable due to failure of OSAM system
#ifndef BNS_IGNORE_MEM_ERROR
	if(bns_chk_sdram()!=BNS_SDRAM_OK){
		printk("SLM: memory error in open()\n");
		return -EIO;
		}
#endif
	
/* to allow double open during debugging -matt 						*/
/* should be clever here and allow overrides with certain perms but the code is good so *
 * this is not required- better compile for hardreset					*/

#ifdef ALLOW_HARD_RESET
        if(slm.locked==TRUE) return -EBUSY;
#endif
        slm.locked=TRUE;
        MOD_INC_USE_COUNT;
        PRINT_DEBUG("slm:open:done\n");
        return 0;
        }

/****************************************************************************************/
/* RELEASE										*/
/* As open hardly does anything this does little as well				*/

static int slm_release(struct inode *inode, struct file *file) {
        MOD_DEC_USE_COUNT;
/* release the device */
        slm.locked=FALSE;
	PRINT_DEBUG("slm:release:done\n");
        return 0;
        }


/****************************************************************************************/
/* SLM UTILITIES									*/
/* a collection of little functions to make life easier					*
 * should be in another file but I want to inline many of these- matt Oct 2001		*/

/* bns select image - push an image onto the card					*/
/* this can ONLY run at safe times- when DPEN is enabled (DPEN high)			*/

static inline void bns_select_image(u32 image,u32 periods, u32 lasera, u32 laserb){
	/* write the critical registers, address RRDY check BURP			*/
	slm.cur_image=image;
	/* these may require a specific sequence according to one source file from BNS	* 
	 * but kipp says any sequence- however all must be written to avoid RRDY error	*/

	set_image(slm.mem_image,slm.cur_image);
	writel(periods,	slm.io+BNS_PERIODS);
	writel(lasera,	slm.io+BNS_LASERA);
	writel(laserb,	slm.io+BNS_LASERB);
	writel(0,	slm.io+BNS_RRDY);
	}	

static inline int bns_check_RRDY(void){
	if(readl(slm.io+BNS_STATUS)&BNS_STAT_RRDY_ERR){
		/* write CLR_BURP to clear error and return error	*/
		writel(0,slm.io+BNS_CLRBURP);
		return -BNS_STAT_RRDY_ERR;
		}
	return 0;
	}	

/********************************************************************************/
/* Power management- "init_board" "power_on" and "power off"			*
 * These need to grow in sophisiscation and set power level flags- matt		*/

/* INIT_BOARDS									*/
/* turns stuff on and off according to address map v3.3				*/
/* (1) 	set RST_2PORT_CNTR high							*
 * (2) 	wait 8ms (call it 10ms/HZ= 1 jiffy x86 (10ms) - matt)			*
 * (3) 	clear RST_2PORT_CNTR							*
 * (4) 	wait 8ms (1 jiffy)							*
 * (5) 	Set DPEN								*
 * (6) 	wait 15ms (2 jiffies)							*
 * (7) 	clear DPEN								*
sds/matt new delay 15ms here to prevent banding problem 5/4/4
 * (8) 	set RST_2PORT_CNTR high							*
 * (9) 	wait 8ms (1 jiffy)							*
 * (10) clear RST_2PORT_CNTR 							*
 * (11) wait 8ms								*
 * (12) set DPEN								*
 * nothing else spec'd, I will assume wait 15 ms, clear DPEN- matt		*/
static int bns_init_board(void){
	up_cntrl(BNS_CNTL_RST_2PORT_CNTR);
	bns_ms_sleep(8);
	down_cntrl(BNS_CNTL_RST_2PORT_CNTR);
	bns_ms_sleep(8);
	up_cntrl(BNS_CNTL_DPEN);
	bns_ms_sleep(15);
	down_cntrl(BNS_CNTL_DPEN);
	/* no delay spec'd here!*/
	/* sds gonna try adding a delay to get rid of banding bug */
	/* sds THIS APPEARS TO BE NECESSARY to get rid of the occasional */
	/* (~1 in 10) chance of a corrupted image ("banding problem")  5 Mar 04 */
	bns_ms_sleep(15); /*15ms chosen semi-randomly, may not need to be that long*/
	up_cntrl(BNS_CNTL_RST_2PORT_CNTR);
	bns_ms_sleep(8);
	down_cntrl(BNS_CNTL_RST_2PORT_CNTR);
	bns_ms_sleep(8);
	up_cntrl(BNS_CNTL_DPEN);

	/* now according to BNS that is it- but I don't want to leave DPEN open */
	/* so I will sleep at least one image period- give it time to DPEN	*/
	//printk("Sleeping for %d ms",(slm.periods*165)/1000);
	// consider removing this.... matt
	bns_ms_sleep((slm.periods*165)/1000);
	down_cntrl(BNS_CNTL_DPEN);

/* botch the return at present /  return slm.power_status (updated) in future- matt	*/
	return TRUE;
	}

/* POWER_ON									*/
/* according to BNS								*
 * (1) set DAC_PWR_CNTRL							*
 * (2) wait on  2PORT_INIT (presumably some time out here? -matt)		*
 *	<takes ~ 200ms (address map v3.3)- give it 1 second- matt>		*
 * (3) set OPAM_PWR_CNTRL							*
 * (4) wait 10ms								*
 * (5) set SLM_PWR_CNTRL 							*
 * nothing else spec'd- is SLM_PWR_CNTRL instant?				*
 */

/* return BNS_PWR_ERR if 2port does not come up "in a reasonable time"		*/
/* I will do this by polling 	2PORT_INIT and sleeping- matt			*/

static int bns_power_on(void){
int i;
u32 stat;
	PRINT_DEBUG("power on called\n");
	slm.power_status=BNS_POWER_IS_OFF;
	up_cntrl(BNS_CNTL_DAC_PWR_CNTRL);  
	for(i=0;i<BNS_PWR_TIMEOUT_LOOPS;i++){
		bns_ms_sleep(10);
		stat=readl(slm.io+BNS_STATUS);
		if(stat&BNS_STAT_2PORT_INIT)break;
		}
	printk("DAC Power took %d ms to come up\n",i*10);
	if(i>=BNS_PWR_TIMEOUT_LOOPS){
		printk("***slm: bns_power_on: 2port failed to power up\n");
		return BNS_PWR_ERR;
		}

	up_cntrl(BNS_CNTL_OPAMP_PWR_CNTL);
	bns_ms_sleep(10);

	up_cntrl(BNS_CNTL_SLM_PWR_CNTL);
	
	/* no final sleep was spec'd- I added it here for sanity- matt		*/
	bns_ms_sleep(10);
	
	stat=readl(slm.io+BNS_STATUS);
	if(stat&BNS_STAT_2PORT_INIT) 	slm.power_status|=BNS_POWER_DAC;
	slm.power_status|=BNS_POWER_OPAMP; /* no flag for this-matt */
	if(stat&BNS_STAT_V_STATUS) 	slm.power_status|=BNS_POWER_VOK;
	if(stat&BNS_STAT_SLM_PWR_OK) 	slm.power_status|=BNS_POWER_SLM;
	if(stat&=BNS_STAT_DAC_PWR_ERR){
		printk("BNS_STAT_DAC_PWR_ERR\n");
		}
	if(stat&=BNS_STAT_ANA_PWR_ERR){
		printk("BNS_STAT_ANA_PWR_ERR\n");
		}

/* now then now then why this? well there is <another> hw bug to get around	*
 * according to the PLXDEBUG instructions "there is a glitch" which can result	*
 * in "banded" SLM output if the counters were init'd before the SLM		*
 * so to get around this init them after the power has stablised this the job	*
 * that is done in bns_init_board- matt nov 2001				*/
/* we may need a final, longer sleep before this though				*/

	bns_init_board();

/* as a matter of policy force the user to request an image to be display	*/	
/* (unless we want to put a logo / test pattern here matt)			*/
	bns_enable_write();
	return slm.power_status;
	}

/* POWER_OFF									*/
/* according to BNS								*
 * (1) clear SLM_PWR_CNTRL							*
 * (2) Wait 200ms (this is a <long> time- matt					*
 * (3) clear OPAMP_PWR_CNTRL							*
 * (4) wait 200ms								*
 * (5) clear DAC_PWR_CNTRL							*
 * again not time delay spec'd here	-reckon on 200ms ? matt			*/

static int bns_power_off(void){
u32 stat;
	/* disable ints / disable DPEN */
	bns_enable_write();
	PRINT_DEBUG("power off called\n");
	down_cntrl(BNS_CNTL_SLM_PWR_CNTL);
	bns_ms_sleep(200);

	down_cntrl(BNS_CNTL_OPAMP_PWR_CNTL);
	bns_ms_sleep(200);
	
	down_cntrl(BNS_CNTL_DAC_PWR_CNTRL);
	
	/* again no final sleep spec'd but I put it here for sanity */
	bns_ms_sleep(200);
	
	/* no indiaction that this fails- but we should probably check for 	*
	 * symmetry with power_on -matt						*/ 
	slm.power_status=BNS_POWER_IS_OFF;
	stat=readl(slm.io+BNS_STATUS);
	if(stat&BNS_STAT_2PORT_INIT) 	slm.power_status|=BNS_POWER_DAC;
	/* BNS_POWER_OPAMP; no flag for this - its off if you say so matt	*/
	if(stat&BNS_STAT_V_STATUS) 	slm.power_status|=BNS_POWER_VOK;
	if(stat&BNS_STAT_SLM_PWR_OK) 	slm.power_status|=BNS_POWER_SLM;
	return slm.power_status;
	}


/* enable / disable write - disable / enable DPEN and ints during writing to card memory	*/
/* this <should> be ok without this by design 						*
 * BUT Kipp said it crashes the computer 10% of the time				*/
/* ints must be stopped or we may update the image pointer and require the DPEN		*/
/* if we are not in SLM_LIST inturrpts have no business being enabled any way		*/

static inline void bns_enable_write(void){
	//if(slm.image_mode==BNS_SLM_LIST)bns_disable_ints();
	/* we can always safely disable ints 						*/
	bns_disable_ints();
	bns_disable_DPEN();
	}		

static inline void bns_disable_write(void){
	/* only enable ints if we are going through the list				*/
	if(slm.image_mode==BNS_SLM_LIST){
		bns_enable_ints();
		}
	bns_enable_DPEN();
	}		

/* enable / disable 2port- allow / disallow transfers from card to port			*/
static inline void bns_disable_DPEN(void){
	down_cntrl(BNS_CNTL_DPEN);
// XXXXXXXXX clean here
	switch(slm.DPEN_mode){
		case BNS_DPEN_MODE_BUSY:
			udelay(slm.DPEN_delay);
			break;
		case BNS_DPEN_MODE_SLEEP:
		/* catch other modes safely */
		case BNS_DPEN_MODE_PROBE:
		default:
			bns_ms_sleep(slm.DPEN_delay/1000);
		}
/* the DPEN is only FALSE after the delay-----						*/
	slm.DPEN_enable=FALSE;
	}

static inline void bns_enable_DPEN(void){
	up_cntrl(BNS_CNTL_DPEN);
/* the DPEN is TRUE immediately								*/
	slm.DPEN_enable=TRUE;
	}

/* check SDRAM programming								*/
static inline int bns_chk_sdram(void){
u32 s0,s1,s2,s3,temp;
//int i;
	s0=readl(slm.io+BNS_SDRAM0);
	s1=readl(slm.io+BNS_SDRAM1);
	s2=readl(slm.io+BNS_SDRAM2);
	s3=readl(slm.io+BNS_SDRAM3);

/* On test system mk 1 memory errors 0x89 were found and these were harmless so ignore	*/ 
#ifdef BNS_0X89_SDRAM_ERROR
	if(s2==0x89){
		slm.mem_err0x89++;
		s2=0x9;
		}
#endif
	temp=(s0|s1|s2|s3);
/* On test system mk 1 the SDRAM controllers could become corrupt- a provision for 	*
 * reprogramming the SDARM FPGAs was enabled the next chunk of code takes care of it	*
 * if the board does not have this provision the behaviour of this code is UNDEFINED	*/
#ifdef BNS_SDRAM_RESET_ENABLED
	if(temp!=BNS_SDRAM_OK){
		printk("%x:%x:%x:%x\n",s0,s1,s2,s3);
		printk("SLM: SDRAM failed trying to fix\n");
		/* reprogram SDRAM controllers						*/
		up_cntrl(BNS_CNTL_SDRAM_PROG);
		/* perform a long busy wait - maybe schedule here-matt?			*/
		udelay(1000);	
		down_cntrl(BNS_CNTL_SDRAM_PROG);
		slm.mem_errs++;
		/* according to Kipp this is not known to fail				*
		 * What would I do otherwise?						*
		 * if it fails then the computer "locks" on write (SDRAM accepts no 	*
		 * memory and the PCI bus chokes)					*/
		/* I'll leave it- getting out of this error is not pretty- if we have   *
		 * trouble I will trap this condition just in case- matt oct 2001	*/
		for(i=0;i<200;i++){
			s0=readl(slm.io+BNS_SDRAM0);
			s1=readl(slm.io+BNS_SDRAM1);
			s2=readl(slm.io+BNS_SDRAM2);
			s3=readl(slm.io+BNS_SDRAM3);
			temp=(s0|s1|s2|s3);
			bns_ms_sleep(10);	
			if(temp==BNS_SDRAM_OK){
				break;
				}
			}
			printk("%d:%x:%x:%x:%x\n",i,s0,s1,s2,s3);
		
		}
#endif
	return temp;
	}

/* delay by going to sleep- note that the smallest delay is 1000/HZ ms	(10ms typically)*/

static void bns_ms_sleep(int ms){
int jif;
	jif=1+ms*HZ/1000; 	/* round up 	*/
	if(jif<1)jif=1;		/* not required	iff ms always +ve */
	current->state=TASK_INTERRUPTIBLE;
	schedule_timeout(jif);
	}


/* bns enable/disable ints								*/
static inline void bns_enable_ints(void){
	up_cntrl(BNS_CNTL_INTEN);
	}
	
static inline void bns_disable_ints(void){
	down_cntrl(BNS_CNTL_INTEN);
	}

/* an init function removed to here to improve the readability in init			*/
static int bns_init_sdram(void){
u32 s0,s1,s2,s3;
	bns_chk_sdram();/* a hack to prevent previous fails getting in the way */
        s0=readl(slm.io+BNS_SDRAM0);
        s1=readl(slm.io+BNS_SDRAM1);
        s2=readl(slm.io+BNS_SDRAM2);
        s3=readl(slm.io+BNS_SDRAM3);
        if((s0!=s1)||(s1!=s2)||(s2!=s3)){
                /* this is trouble- the reset sdrams have the wrong ids */
                printk("slm: init: init_SDRAM controller ids error (%x,%x,%x,%x)\n",s0,s1,s2,s3);
                return -ENODEV;
                }
        if(s0==BNS_SDRAM_FAIL){
                printk("slm: init: SDRAM fail, attempting reset\n");
                bns_chk_sdram();
                s0=readl(slm.io+BNS_SDRAM0);
                s1=readl(slm.io+BNS_SDRAM1);
                s2=readl(slm.io+BNS_SDRAM2);
                s3=readl(slm.io+BNS_SDRAM3);
                if((s0!=s1)||(s1!=s2)||(s2!=s3)){
                        /* this is trouble- the reset sdrams have the wrong ids */
                        printk("slm: init: init_SDRAM: after reset: ids error (%x,%x,%x,%x)\n",s0,s1,s2,s3);
                        return -ENODEV;
                        }
                }
        if(s0==BNS_SDRAM_FAIL){
                printk("slm: sdram reset failed with value BNS_SDRAM_FAIL\n");
		printk("slm: Perhaps the 2PORT is not connected?\n");
		
		// added by matt nov 2003 
		// XXXX disabled to allow OSAM fail debug
		return -ENODEV;
                }
        slm.sdram_id=s0;
	return 0;
	}

static int bns_probe_memory_board(void){
u32 temp;
/* find out how much memory is on board                                 */
        temp=readl(slm.io+BNS_STATUS);
        slm.mem_board_id=BNS_GET_MEM_BOARD_ID(temp);
        printk("slm: mem board id probed, id=%x, me id=%x ",slm.mem_board_id,slm.mem_board_id&BNS_MEM_ID_MASK);
        switch(slm.mem_board_id&BNS_MEM_ID_MASK){
                case BNS_MEM_1024:
                        printk("contains 1024 images\n");
                        slm.max_images=1024;
                        break;
                case BNS_MEM_2048:
                        printk("contains 2048 images\n");
                        slm.max_images=2048;
                        break;
                case BNS_MEM_512:
                        printk("contains 512 images\n");
                        slm.max_images=512;
                        break;
                case BNS_MEM_UNDEFINED:
                default:
                        printk("undefined memory ID, cleaning up and exiting\n");
                        return -ENODEV;
                }
	return 0;
	}

/* The EPI is an external interface on the card- 			*
 *				it does not concern the SLM at present	*/
static int bns_probe_epi(void){
u32 temp;
        temp=readl(slm.io+BNS_STATUS);
        slm.epi_id=BNS_GET_EPI_ID(temp);
        printk("slm: epi id probed, status=%x, id=%x, ",temp,slm.epi_id);
        switch(slm.epi_id){
                case BNS_EPI_PRESENT_NOT_INIT:
                        printk("epi present and uninit'd (good)\n");
                        break;
                case BNS_EPI_NOT_PRESENT:
                        printk("no epi present (ok) ignoring\n");
                        break;
                case BNS_EPI_PRESENT_SUNDANCE:
                        printk("epi present and init'd, id=\"modified Sundance SMT-335\"\n");
                        break;
                default:
                        printk("epi present and init'd, id unknown\n");
                }
	return 0;
	}

static void bns_probe_plx_int(void){
u32 temp;
	temp=readl(slm.plx_chip+PLX_INTERRUPT);
	printk("slm: probing PLX interrupt register\n");
	printk("slm: register PLX_INTERRUPT=%x\n",temp);
	
	printk("slm: PCI interrupts are ");
	if(temp&PLX_INT_PCI)
		printk("enabled ");
	else
		printk("disabled ");
	if(temp&PLX_INT_SOFT_INT)
		printk("(soft on, problem?)\n");
	else
		printk("(soft off- good)\n");
	
	printk("slm: ISA mode is ");
	if(temp&PLX_INT_PCI)
		printk("enabled ");
	else
		printk("disabled ");
	
	printk("\nslm: local int 1 is ");
	if(temp&PLX_INT_LINT1_ON)
		printk("on, ");
	else
		printk("off, ");
	if(temp&PLX_INT_LINT1_POL)
		printk("active high, ");
	else
		printk("active low, ");
	if(temp&PLX_INT_LINT1_SEL)
		printk("edge trig'd and ");
	else
		printk("level trig'd and ");
	if(temp&PLX_INT_LINT1_STAT)
		printk("active and");
	else
		printk("inactive and");
	if(temp&PLX_INT_LET_CLR1)
		printk(" ET cleared\n");
	else
		printk(" ET unclear\n");


	printk("slm: local int 2 is ");
	if(temp&PLX_INT_LINT2_ON)
		printk("on, ");
	else
		printk("off, ");
	if(temp&PLX_INT_LINT2_POL)
		printk("active high, ");
	else
		printk("active low, ");
	if(temp&PLX_INT_LINT2_SEL)
		printk("edge trig'd and ");
	else
		printk("level trig'd and ");
	if(temp&PLX_INT_LINT2_STAT)
		printk("active and");
	else
		printk("inactive and");
	if(temp&PLX_INT_LET_CLR2)
		printk(" ET cleared\n");
	else
		printk(" ET unclear\n");
	}

static inline void set_cntrl(u32 val){
	control=val;
	writel(control,slm.io+BNS_CONTROL);
	}

static inline void up_cntrl(u32 val){
	control|=val;
	writel(control,slm.io+BNS_CONTROL);
	}

static inline void down_cntrl(u32 val){
	control&=~val;
	writel(control,slm.io+BNS_CONTROL);
	}

/* Control were the various images go						*/
/* the LSB 16 bits (10 used) point to the image to be transferred to the SLM	*/
/* the MSB 16 bits point to the image memory slot that the host writes to	*/
static inline void set_image(int mem_image,int slm_image){
u32	image;
	slm.cur_image=mem_image;
	image=((mem_image<<16)&BNS_MEM_IMAGE)|(slm_image&BNS_DP_IMAGE);
	writel(image,slm.io+BNS_IMAGE);
	}





